# -*- coding: utf-8 -*-
"""
       .Specific exceptions for pyCSAMT.
    .Created on Wed Nov 18 17:17:13 2020

@author: @Daniel03

"""

#=================================================================
class pyCSAMTError_occam2d_plot(Exception): 
    pass 
class pyCSAMTError_occam2d(Exception): 
    pass 
class pyCSAMTError_occam2d_iter2dat(Exception): 
    pass

class pyCSAMTError_profile(Exception): 
    pass 

class pyCSAMTError_J(Exception):
    pass

class pyCSAMTError_site(Exception): 
    pass 

class MTpyError_Z(Exception):
    pass 

class MTpyError_inputarguments(Exception): 
    pass 

class pyCSAMTError_EDI(Exception):
    pass 

class pyCSAMTError_location(Exception): 
    pass 
class pyCSAMTError_well_building(Exception): 
    pass 

class pyCSAMTError_float(Exception):
    pass
class pyCSAMTError_frequency(Exception):
    pass

class pyCSAMTError_inputarguments(Exception):
    pass

class pyCSAMTError_Header(Exception):
    pass 

class pyCSAMTError_value(Exception):
    pass
class pyCSAMTError_azimuth(Exception): 
    pass 

class pyCSAMTError_ts_data(Exception):
    pass


class pyCSAMTError_config_file(Exception):
    pass


class pyCSAMTError_file_handling(Exception):
    pass


class pyCSAMTError_avg_file(Exception):
    pass

class pyCSAMTError_AvgData(Exception):
    pass


class pyCSAMTError_AVG(Exception):
    pass

class pyCSAMTError_plot_tip(Exception):
    pass 
class pyCSAMTError_plot(Exception): 
    pass 

class pyCSAMTError_Z(Exception):
    pass


class pyCSAMTError_Emag(Exception):
    pass


class pyCSAMTError_Hmag(Exception):
    pass

class pyCSAMTError_Ephz(Exception):
    pass

class pyCSAMTError_parameter_number(Exception):
    pass

class pyCSAMTError_processing(Exception):
    pass

class pyCSAMTError_module_import(Exception):
    pass

class pyCSAMTError_Phase(Exception):
    pass 

class pyCSAMTError_rho(Exception):
    pass

class pyCSAMTError_station(Exception):
    pass

class pyCSAMTError_structural(Exception): 
    pass

class pyCSAMTError_strata(Exception): 
    pass

class pyCSAMTError_plot_geoinputargument(Exception): 
    pass
class pyCSAMTError_geodrill_inputarguments(Exception): 
    pass

class pyCSAMTError_SQL(Exception): 
    pass
class pyCSAMTError_SQL_manager(Exception): 
    pass
class pyCSAMTError_SQL_interface(Exception): 
    pass
class pyCSAMTError_SQL_geoDataBase(Exception): 
    pass

class pyCSAMTError_SQL_update_geoinformation(Exception): 
    pass




